from dataclasses import dataclass

import aiohttp

from .types import Request, RequestFailed, Response


@dataclass(frozen=True)
class AIOHTTP:
    session: aiohttp.ClientSession

    async def __call__(self, request: Request) -> Response:
        try:
            async with self.session.request(
                request.method, request.url, headers=request.headers, data=request.body
            ) as response:
                return Response(response.status, await response.read())
        except aiohttp.ClientError as exc:
            raise RequestFailed(exc)



import asyncio
from dataclasses import dataclass
from typing import Dict, cast

import httpx

from .types import Request, RequestFailed, Response


@dataclass(frozen=True)
class HTTPX:
    client: httpx.AsyncClient

    async def __call__(self, request: Request) -> Response:
        try:
            response = await self.client.request(
                method=request.method,
                url=request.url,
                # httpx is coded with no_implicit_optional=False, we use strict=True
                headers=cast(Dict[str, str], request.headers),
                content=cast(bytes, request.body),
            )
            return Response(response.status_code, await response.aread())
        except httpx.TimeoutException:
            raise asyncio.TimeoutError()
        except httpx.HTTPError as exc:
            raise RequestFailed(exc)



from __future__ import annotations

import abc
import asyncio
import datetime
import random
import time
from dataclasses import dataclass
from enum import Enum, unique
from itertools import count
from typing import (
    Any,
    AsyncIterable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Union,
    cast,
)

from .expressions import Parameters, ProjectionExpression
from .types import (
    EncodedGlobalSecondaryIndex,
    EncodedKeySchema,
    EncodedLocalSecondaryIndex,
    EncodedPayPerRequest,
    EncodedProjection,
    EncodedStreamSpecification,
    EncodedThroughput,
    Item,
    Seconds,
    TableName,
)
from .utils import py2dy


@unique
class TimeToLiveStatus(Enum):
    enabling = "ENABLING"
    disabling = "DISABLING"
    enabled = "ENABLED"
    disabled = "DISABLED"


@dataclass(frozen=True)
class TimeToLiveDescription:
    table: str
    attribute: str
    status: TimeToLiveStatus


@dataclass(frozen=True)
class Throughput:
    read: int
    write: int

    def encode(self) -> EncodedThroughput:
        return {
            "ProvisionedThroughput": {
                "ReadCapacityUnits": self.read,
                "WriteCapacityUnits": self.write,
            }
        }


@dataclass(frozen=True)
class PayPerRequest:
    MODE = "PAY_PER_REQUEST"

    def encode(self) -> EncodedPayPerRequest:
        return {"BillingMode": PayPerRequest.MODE}


ThroughputType = Union[Throughput, PayPerRequest]


class KeyType(Enum):
    string = "S"
    number = "N"
    binary = "B"


@dataclass(frozen=True)
class KeySpec:
    name: str
    type: KeyType


@dataclass(frozen=True)
class KeySchema:
    hash_key: KeySpec
    range_key: Optional[KeySpec] = None

    def __iter__(self) -> Iterator[KeySpec]:
        yield self.hash_key

        if self.range_key:
            yield self.range_key

    def to_attributes(self) -> Dict[str, str]:
        return {key.name: key.type.value for key in self}

    def encode(self) -> List[EncodedKeySchema]:
        return [
            {"AttributeName": key.name, "KeyType": key_type}
            for key, key_type in zip(self, ["HASH", "RANGE"])
        ]


class ProjectionType(Enum):
    all = "ALL"
    keys_only = "KEYS_ONLY"
    include = "INCLUDE"


@dataclass(frozen=True)
class Projection:
    type: ProjectionType
    attrs: Optional[List[str]] = None

    def encode(self) -> EncodedProjection:
        encoded: EncodedProjection = {"ProjectionType": self.type.value}
        if self.attrs:
            encoded["NonKeyAttributes"] = self.attrs
        return encoded


@dataclass(frozen=True)
class LocalSecondaryIndex:
    name: str
    schema: KeySchema
    projection: Projection

    def encode(self) -> EncodedLocalSecondaryIndex:
        return {
            "IndexName": self.name,
            "KeySchema": self.schema.encode(),
            "Projection": self.projection.encode(),
        }


@dataclass(frozen=True)
class GlobalSecondaryIndex(LocalSecondaryIndex):
    throughput: Optional[Throughput]

    def encode(self) -> EncodedGlobalSecondaryIndex:
        # mypy really does not like an if/else assignment where the branches have
        # different types, so we need to do some ridiculous casting here.
        # The outer cst is due to https://github.com/python/mypy/issues/4122
        # The two inner casts to Any are because of the if/else with different types
        return cast(
            EncodedGlobalSecondaryIndex,
            {
                **super().encode(),
                **(
                    cast(Any, self.throughput.encode())
                    if self.throughput
                    else cast(Any, {})
                ),
            },
        )


class StreamViewType(Enum):
    keys_only = "KEYS_ONLY"
    new_image = "NEW_IMAGE"
    old_image = "OLD_IMAGE"
    new_and_old_images = "NEW_AND_OLD_IMAGES"


@dataclass(frozen=True)
class StreamSpecification:
    enabled: bool = False
    view_type: StreamViewType = StreamViewType.new_and_old_images

    def encode(self) -> EncodedStreamSpecification:
        spec: EncodedStreamSpecification = {"StreamEnabled": self.enabled}
        if self.enabled:
            spec["StreamViewType"] = self.view_type.value
        return spec


class ReturnValues(Enum):
    none = "NONE"
    all_old = "ALL_OLD"
    updated_old = "UPDATED_OLD"
    all_new = "ALL_NEW"
    updated_new = "UPDATED_NEW"


class TableStatus(Enum):
    creating = "CREATING"
    updating = "UPDATING"
    deleting = "DELETING"
    active = "ACTIVE"


@dataclass(frozen=True)
class TableDescription:
    attributes: Optional[Dict[str, KeyType]]
    created: Optional[datetime.datetime]
    item_count: Optional[int]
    key_schema: Optional[KeySchema]
    throughput: Optional[ThroughputType]
    status: TableStatus

    @classmethod
    def from_response(cls, description: Dict[str, Any]) -> "TableDescription":
        attributes: Optional[Dict[str, KeyType]]
        if "AttributeDefinitions" in description:
            attributes = {
                attribute["AttributeName"]: KeyType(attribute["AttributeType"])
                for attribute in description["AttributeDefinitions"]
            }
        else:
            attributes = None
        creation_time: Optional[datetime.datetime]
        if "CreationDateTime" in description:
            creation_time = datetime.datetime.fromtimestamp(
                description["CreationDateTime"], datetime.timezone.utc
            )
        else:
            creation_time = None
        key_schema: Optional[KeySchema]
        if attributes and "KeySchema" in description:
            key_schema = KeySchema(
                *[
                    KeySpec(
                        name=key["AttributeName"], type=attributes[key["AttributeName"]]
                    )
                    for key in description["KeySchema"]
                ]
            )
        else:
            key_schema = None
        throughput: Optional[ThroughputType]
        if (
            "BillingModeSummary" in description
            and description["BillingModeSummary"]["BillingMode"] == PayPerRequest.MODE
        ):
            throughput = PayPerRequest()
        elif "ProvisionedThroughput" in description:
            throughput = Throughput(
                read=description["ProvisionedThroughput"]["ReadCapacityUnits"],
                write=description["ProvisionedThroughput"]["WriteCapacityUnits"],
            )
        else:
            throughput = None
        return TableDescription(
            attributes=attributes,
            created=creation_time,
            item_count=description.get("ItemCount", None),
            key_schema=key_schema,
            throughput=throughput,
            status=TableStatus(description["TableStatus"]),
        )


class Select(Enum):
    all_attributes = "ALL_ATTRIBUTES"
    all_projected_attributes = "ALL_PROJECTED_ATTRIBUTES"
    count = "COUNT"
    specific_attributes = "SPECIFIC_ATTRIBUTES"


class RetryTimeout(Exception):
    pass


# https://github.com/python/mypy/issues/5374
@dataclass(frozen=True)
class MyPyWorkaroundRetryConfigBase:
    time_limit_secs: Seconds = 60


class RetryConfig(MyPyWorkaroundRetryConfigBase, metaclass=abc.ABCMeta):
    @classmethod
    def default(cls) -> RetryConfig:
        """
        Default RetryConfig to be used as a throttle config for Client
        """
        return ExponentialBackoffRetry()

    @classmethod
    def default_wait_config(cls) -> RetryConfig:
        """
        Default RetryConfig to be used as wait config for table level operations.
        """
        return ExponentialBackoffRetry(time_limit_secs=500)

    @abc.abstractmethod
    def delays(self) -> Iterable[Seconds]:
        """
        Custom RetryConfig classes must implement this method. It should return
        an iterable yielding numbers of seconds indicating the delay before the
        next attempt is made.
        """
        raise NotImplementedError()

    async def attempts(self) -> AsyncIterable[None]:
        deadline = time.monotonic() + self.time_limit_secs
        for delay in self.delays():
            yield
            if time.monotonic() > deadline:
                raise RetryTimeout()
            await asyncio.sleep(delay)


@dataclass(frozen=True)
class StaticDelayRetry(RetryConfig):
    delay: Seconds = 1

    def delays(self) -> Iterable[Seconds]:
        while True:
            yield self.delay


@dataclass(frozen=True)
class DecorelatedJitterRetry(RetryConfig):
    base_delay_secs: Seconds = 0.05
    max_delay_secs: Seconds = 1

    def delays(self) -> Iterable[Seconds]:
        current_delay_secs = self.base_delay_secs
        while True:
            current_delay_secs = min(
                self.max_delay_secs,
                random.uniform(self.base_delay_secs, current_delay_secs * 3),
            )
            yield current_delay_secs


@dataclass(frozen=True)
class ExponentialBackoffRetry(RetryConfig):
    base_delay_secs: Seconds = 2
    max_delay_secs: Seconds = 20

    def delays(self) -> Iterable[Seconds]:
        for attempt in count():
            yield min(
                random.random() * (self.base_delay_secs**attempt), self.max_delay_secs
            )


@dataclass(frozen=True)
class Page:
    items: List[Item]
    last_evaluated_key: Optional[Dict[str, Any]]

    @property
    def is_last_page(self) -> bool:
        return self.last_evaluated_key is None


@dataclass(frozen=True)
class BatchGetRequest:
    keys: List[Item]
    projection: Optional[ProjectionExpression] = None
    consistent_read: bool = False

    def to_request_payload(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "Keys": [py2dy(key) for key in self.keys],
            "ConsistentRead": self.consistent_read,
        }
        if self.projection:
            params = Parameters()
            payload["ProjectionExpression"] = self.projection.encode(params)
            payload.update(params.to_request_payload())

        return payload


@dataclass(frozen=True)
class BatchGetResponse:
    items: Dict[TableName, List[Item]]
    unprocessed_keys: Dict[TableName, List[Item]]


@dataclass(frozen=True)
class BatchWriteRequest:
    keys_to_delete: Optional[List[Item]] = None
    items_to_put: Optional[List[Item]] = None

    def to_request_payload(self) -> List[Dict[str, Any]]:
        payload: List[Dict[str, Any]] = []
        if self.keys_to_delete:
            payload.extend(
                {"DeleteRequest": {"Key": py2dy(key)}} for key in self.keys_to_delete
            )
        if self.items_to_put:
            payload.extend(
                {"PutRequest": {"Item": py2dy(item)}} for item in self.items_to_put
            )
        return payload


@dataclass(frozen=True)
class BatchWriteResult:
    undeleted_keys: List[Item]
    unput_items: List[Item]



from __future__ import annotations

import abc
import asyncio
import configparser
import datetime
import json
import os
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import (
    Callable,
    Coroutine,
    Generic,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
)

from yarl import URL

from .http.types import HttpImplementation, Request, RequestFailed
from .types import Timeout
from .utils import logger, parse_amazon_timestamp


@dataclass(frozen=True)
class Key:
    id: str
    secret: str = field(repr=False)
    token: Optional[str] = field(repr=False, default=None)


class Credentials(metaclass=abc.ABCMeta):
    @classmethod
    def auto(cls) -> ChainCredentials:
        """
        Return the default credentials loader chain.
        """
        return ChainCredentials(
            candidates=[
                EnvironmentCredentials(),
                FileCredentials(),
                ContainerMetadataCredentials(),
                InstanceMetadataCredentialsV2(),
                InstanceMetadataCredentialsV1(),
            ]
        )

    @abc.abstractmethod
    async def get_key(self, http: HttpImplementation) -> Optional[Key]:
        """
        Return a Key if one could be found.
        """

    @abc.abstractmethod
    def invalidate(self) -> bool:
        """
        Invalidate the credentials if possible and return whether credentials
        got invalidated or not.
        """

    @abc.abstractmethod
    def is_disabled(self) -> bool:
        """
        Indicate if this credentials provider is disabled. Used by ChainCredentials
        to ignore providers that won't ever find a key.

        If the status could change over the lifetime of a program, return True.
        """


class Disabled(Exception):
    pass


@dataclass(frozen=True)
class StaticCredentials(Credentials):
    """
    Static credentials provided in Python.
    """

    key: Optional[Key]

    async def get_key(self, http: HttpImplementation) -> Optional[Key]:
        return self.key

    def invalidate(self) -> bool:
        return False

    def is_disabled(self) -> bool:
        return False


class ChainCredentials(Credentials):
    """
    Chains multiple credentials providers together, trying them
    in order. Returns the first key found. Exceptions are suppressed.

    Once a credentials provider returns a key, only that provider will
    be used in subsequent calls.
    """

    _candidates: Sequence[Credentials]
    _chosen: Credentials | None

    def __init__(self, candidates: Sequence[Credentials]) -> None:
        self._candidates = [
            candidate for candidate in candidates if not candidate.is_disabled()
        ]
        self._chosen = None

    async def get_key(self, http: HttpImplementation) -> Optional[Key]:
        if self._chosen is not None:
            return await self._chosen.get_key(http)
        for candidate in self._candidates:
            try:
                key = await candidate.get_key(http)
            except:
                logger.exception("Candidate %r failed", candidate)
                continue
            if key is None:
                logger.debug("Candidate %r didn't find a key", candidate)
            else:
                logger.debug("Candidate %r found a key %r", candidate, key)
                self._chosen = candidate
                return key
        return None

    def invalidate(self) -> bool:
        return any(candidate.invalidate() for candidate in self._candidates)

    def is_disabled(self) -> bool:
        return not self._candidates


class EnvironmentCredentials(Credentials):
    """
    Loads the credentials from the environment.
    """

    key: Optional[Key]

    def __init__(self) -> None:
        try:
            self.key = Key(
                os.environ["AWS_ACCESS_KEY_ID"],
                os.environ["AWS_SECRET_ACCESS_KEY"],
                os.environ.get("AWS_SESSION_TOKEN", None),
            )
        except KeyError:
            self.key = None

    async def get_key(self, http: HttpImplementation) -> Optional[Key]:
        return self.key

    def invalidate(self) -> bool:
        return False

    def is_disabled(self) -> bool:
        return self.key is None


# https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html
class FileCredentials(Credentials):
    """
    Loads the credentials from an AWS credentials file
    """

    key: Optional[Key]

    def __init__(
        self,
        *,
        path: Optional[Path] = None,
        profile_name: Optional[str] = None,
    ) -> None:
        if profile_name is None:
            profile_name = cast(str, os.environ.get("AWS_PROFILE", "default"))
        if path is None:
            path = Path.home().joinpath(".aws", "credentials")
        self.key = None
        if not path.is_file():
            return

        parser = configparser.RawConfigParser()
        try:
            parser.read(path)
        except configparser.Error:
            logger.exception("Found credentials file %r but parsing failed", path)
            return

        try:
            profile = parser[profile_name]
        except KeyError:
            logger.exception(
                "Profile %r not found in credentials file %r", profile_name, path
            )
            return

        try:
            self.key = Key(
                id=profile["aws_access_key_id"],
                secret=profile["aws_secret_access_key"],
                token=profile.get("aws_session_token", None),
            )
        except KeyError:
            logger.exception(
                "Profile %r in %r does not contain credentials", profile_name, path
            )

    async def get_key(self, http: HttpImplementation) -> Optional[Key]:
        return self.key

    def invalidate(self) -> bool:
        return False

    def is_disabled(self) -> bool:
        return self.key is None


class Refresh(Enum):
    required = auto()
    soon = auto()
    not_required = auto()


EXPIRES_SOON_THRESHOLD = datetime.timedelta(minutes=15)
EXPIRED_THRESHOLD = datetime.timedelta(minutes=10)


class _Unset:
    pass


_UNSET = _Unset()


T = TypeVar("T")
U = TypeVar("U")


@dataclass
class Refreshable(Generic[T]):
    name: str
    should_refresh: Callable[[T], Refresh]
    do_refresh: Callable[[HttpImplementation], Coroutine[None, None, T]]
    _active_refresh_task: Optional[asyncio.Task[T]] = None
    _current: Union[_Unset, T] = _UNSET

    async def get(self, http: HttpImplementation) -> T:
        refresh = self._check_refresh()
        logger.debug("%s refresh status %r", self.name, refresh)
        if refresh is Refresh.required:
            if self._active_refresh_task is None:
                logger.debug("%s starting mandatory refresh", self.name)
                self._active_refresh_task = task = asyncio.create_task(
                    self.do_refresh(http)
                )
                task.add_done_callback(self._clear_active_refresh)
            else:
                logger.debug("%s re-using active refresh", self.name)
            self._current = await self._active_refresh_task
        elif refresh is Refresh.soon:
            if self._active_refresh_task is None:
                logger.debug("%s starting early refresh", self.name)
                self._active_refresh_task = task = asyncio.create_task(
                    self.do_refresh(http)
                )
                task.add_done_callback(self._clear_active_refresh)
                task.add_done_callback(self._set_current)
            else:
                logger.debug("%s already refreshing", self.name)
        assert not isinstance(self._current, _Unset)
        return self._current

    def invalidate(self) -> None:
        self._current = _UNSET

    def _check_refresh(self) -> Refresh:
        if isinstance(self._current, _Unset):
            return Refresh.required
        return self.should_refresh(self._current)

    def _set_current(self, task: asyncio.Task[T]) -> None:
        self._current = task.result()

    def _clear_active_refresh(self, _task: asyncio.Task[T]) -> None:
        self._active_refresh_task = None


@dataclass(frozen=True)
class Metadata:
    key: Key
    expires: datetime.datetime

    def check_refresh(self) -> Refresh:
        return check_refresh(self.expires)


def check_refresh(expires: datetime.datetime) -> Refresh:
    now = datetime.datetime.now(datetime.timezone.utc)
    if now >= expires:
        return Refresh.required
    diff = expires - now
    if diff < EXPIRED_THRESHOLD:
        return Refresh.required
    if diff < EXPIRES_SOON_THRESHOLD:
        return Refresh.soon
    return Refresh.not_required


class MetadataCredentials(Credentials, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def fetch_metadata(self, http: HttpImplementation) -> Metadata:
        pass

    def __post_init__(self) -> None:
        self._refresher: Refreshable[Metadata] = Refreshable(
            f"metadata-credentials-{self.__class__.__name__}",
            Metadata.check_refresh,
            self.fetch_metadata,
        )

    async def get_key(self, http: HttpImplementation) -> Optional[Key]:
        if self.is_disabled():
            logger.debug("%r is disabled", self)
            return None
        metadata = await self._refresher.get(http)
        return metadata.key

    def invalidate(self) -> bool:
        logger.debug("%r invalidated", self)
        self._refresher.invalidate()
        return True


def and_then(thing: Optional[T], mapper: Callable[[T], U]) -> Optional[U]:
    if thing is not None:
        return mapper(thing)
    return thing


# https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-metadata-endpoint.html
@dataclass
class ContainerMetadataCredentials(MetadataCredentials):
    """
    Loads credentials from the ECS container metadata endpoint.
    """

    timeout: Timeout = 2
    max_attempts: int = 3
    base_url: URL = URL("http://169.254.170.2")
    relative_uri: Optional[str] = field(
        default_factory=lambda: os.environ.get(
            "AWS_CONTAINER_CREDENTIALS_RELATIVE_URI", None
        )
    )
    full_uri: Optional[URL] = field(
        default_factory=lambda: and_then(
            os.environ.get("AWS_CONTAINER_CREDENTIALS_FULL_URI", None), URL
        )
    )
    auth_token: Optional[str] = field(
        default_factory=lambda: os.environ.get(
            "AWS_CONTAINER_AUTHORIZATION_TOKEN", None
        )
    )

    @property
    def url(self) -> Optional[URL]:
        if self.relative_uri is not None:
            return self.base_url.with_path(self.relative_uri)
        elif self.full_uri is not None:
            return self.full_uri
        else:
            return None

    def is_disabled(self) -> bool:
        return self.url is None

    async def fetch_metadata(self, http: HttpImplementation) -> Metadata:
        headers = and_then(self.auth_token, lambda token: {"Authorization": token})
        if self.url is None:
            raise Disabled()
        response = await fetch_with_retry_and_timeout(
            http=http,
            max_attempts=self.max_attempts,
            timeout=self.timeout,
            request=Request(
                method="GET", url=str(self.url), headers=headers, body=None
            ),
        )
        data = json.loads(response)
        return Metadata(
            key=Key(
                id=data["AccessKeyId"],
                secret=data["SecretAccessKey"],
                token=data.get("Token", None),
            ),
            expires=parse_amazon_timestamp(data["Expiration"]),
        )


@dataclass
class AuthToken:
    value: str
    expires: datetime.datetime

    def check_refresh(self) -> Refresh:
        return check_refresh(self.expires)


MAX_IMDSV2_AUTH_TOKEN_SESSION_DURATION = 6 * 60 * 60
DEFAULT_IMDSV2_AUTH_TOKEN_SESSION_DURATION = MAX_IMDSV2_AUTH_TOKEN_SESSION_DURATION

# https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instancedata-data-retrieval.html
@dataclass
class InstanceMetadataCredentialsV2(MetadataCredentials):
    """
    Loads credentials from the EC2 instance metadata endpoint using IMDSv2.
    IMDSv2 is considered more secure than IMDSv1, but it may not be available in older AWS SDKs.
    """

    timeout: Timeout = 1
    max_attempts: int = 1
    base_url: URL = URL("http://169.254.169.254")
    disabled: bool = field(
        default_factory=lambda: os.environ.get(
            "AWS_EC2_METADATA_DISABLED", "false"
        ).lower()
        == "true"
    )
    token_session_duration_seconds: int = DEFAULT_IMDSV2_AUTH_TOKEN_SESSION_DURATION
    auth_token: Refreshable[AuthToken] = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.token_session_duration_seconds > MAX_IMDSV2_AUTH_TOKEN_SESSION_DURATION:
            raise ValueError(
                f"IMDS Version 2 auth token session duration must be lower than "
                f"{MAX_IMDSV2_AUTH_TOKEN_SESSION_DURATION}, "
                f"got {self.token_session_duration_seconds}"
            )
        self.auth_token: Refreshable[AuthToken] = Refreshable(
            "imdsv2-token-refresher", AuthToken.check_refresh, self._fetch_token
        )

    def is_disabled(self) -> bool:
        return self.disabled

    async def fetch_metadata(self, http: HttpImplementation) -> Metadata:
        auth_token = await self.auth_token.get(http)
        token_headers = {"X-aws-ec2-metadata-token": auth_token.value}
        role_url = self.base_url.with_path(
            "/latest/meta-data/iam/security-credentials/"
        )
        role = (
            await fetch_with_retry_and_timeout(
                http=http,
                max_attempts=self.max_attempts,
                timeout=self.timeout,
                request=Request(
                    method="GET", url=str(role_url), headers=token_headers, body=None
                ),
            )
        ).decode("utf-8")
        credentials_url = self.base_url.with_path(
            f"/latest/meta-data/iam/security-credentials/{role}"
        )
        raw_credentials = await fetch_with_retry_and_timeout(
            http=http,
            max_attempts=self.max_attempts,
            timeout=self.timeout,
            request=Request(
                method="GET", url=str(credentials_url), headers=token_headers, body=None
            ),
        )
        credentials = json.loads(raw_credentials)
        return Metadata(
            key=Key(
                credentials["AccessKeyId"],
                credentials["SecretAccessKey"],
                credentials.get("Token", None),
            ),
            expires=parse_amazon_timestamp(credentials["Expiration"]),
        )

    async def _fetch_token(
        self,
        http: HttpImplementation,
    ) -> AuthToken:
        token_url = self.base_url.with_path("/latest/api/token/")
        expires = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(
            seconds=self.token_session_duration_seconds
        )
        token = (
            await fetch_with_retry_and_timeout(
                http=http,
                max_attempts=self.max_attempts,
                timeout=self.timeout,
                request=Request(
                    method="PUT",
                    url=str(token_url),
                    headers={
                        "X-aws-ec2-metadata-token-ttl-seconds": str(
                            self.token_session_duration_seconds
                        )
                    },
                    body=None,
                ),
            )
        ).decode("utf-8")
        return AuthToken(token, expires)


@dataclass
class InstanceMetadataCredentialsV1(MetadataCredentials):
    """
    Loads credentials from the EC2 instance metadata endpoint using IMDSv1.
    """

    timeout: Timeout = 1
    max_attempts: int = 1
    base_url: URL = URL("http://169.254.169.254")
    disabled: bool = field(
        default_factory=lambda: os.environ.get(
            "AWS_EC2_METADATA_DISABLED", "false"
        ).lower()
        == "true"
    )

    def is_disabled(self) -> bool:
        return self.disabled

    async def fetch_metadata(self, http: HttpImplementation) -> Metadata:
        role_url = self.base_url.with_path(
            "/latest/meta-data/iam/security-credentials/"
        )
        role = (
            await fetch_with_retry_and_timeout(
                http=http,
                max_attempts=self.max_attempts,
                timeout=self.timeout,
                request=Request(
                    method="GET", url=str(role_url), headers=None, body=None
                ),
            )
        ).decode("utf-8")
        credentials_url = self.base_url.with_path(
            f"/latest/meta-data/iam/security-credentials/{role}"
        )
        raw_credentials = await fetch_with_retry_and_timeout(
            http=http,
            max_attempts=self.max_attempts,
            timeout=self.timeout,
            request=Request(
                method="GET", url=str(credentials_url), headers=None, body=None
            ),
        )
        credentials = json.loads(raw_credentials)
        return Metadata(
            key=Key(
                credentials["AccessKeyId"],
                credentials["SecretAccessKey"],
                credentials.get("Token", None),
            ),
            expires=parse_amazon_timestamp(credentials["Expiration"]),
        )


class TooManyRetries(Exception):
    pass


async def fetch_with_retry_and_timeout(
    *,
    http: HttpImplementation,
    max_attempts: int,
    timeout: Timeout,
    request: Request,
) -> bytes:
    exception: Optional[Exception] = None
    for _ in range(max_attempts):
        try:
            response = await asyncio.wait_for(http(request), timeout)
        except asyncio.TimeoutError:
            logger.debug("timed out talking to metadata service")
            continue
        except RequestFailed as exc:
            logger.debug("request to metadata service failed")
            exception = exc.inner
            continue
        logger.debug(
            "fetched metadata %s (%s bytes)", response.status, len(response.body)
        )
        if response.status == 200:
            return response.body
    if exception is not None:
        raise exception
    raise TooManyRetries()



import decimal
from enum import Enum
from typing import Any, Callable, Dict, List, TypedDict, Union

Timeout = Union[float, int]
Numeric = Union[float, int, decimal.Decimal]

Item = Dict[str, Any]
DynamoItem = Dict[str, Dict[str, Any]]
TableName = str


class ParametersDict(TypedDict, total=False):
    ExpressionAttributeNames: Dict[str, str]
    ExpressionAttributeValues: Dict[str, Dict[str, Any]]


NOTHING = object()

Seconds = Union[float, int]


class AttributeType(Enum):
    string = "S"
    string_set = "SS"
    number = "N"
    number_set = "NS"
    binary = "B"
    binary_set = "BS"
    boolean = "BOOL"
    null = "NULL"
    list = "L"
    map = "M"


class EncodedThroughputData(TypedDict):
    ReadCapacityUnits: int
    WriteCapacityUnits: int


class EncodedThroughput(TypedDict):
    ProvisionedThroughput: EncodedThroughputData


class EncodedPayPerRequest(TypedDict):
    BillingMode: str


class EncodedKeySchema(TypedDict):
    AttributeName: str
    KeyType: str


class EncodedProjectionRequired(TypedDict):
    ProjectionType: str


class EncodedProjection(EncodedProjectionRequired, total=False):
    NonKeyAttributes: List[str]


class EncodedLocalSecondaryIndex(TypedDict):
    IndexName: str
    KeySchema: List[EncodedKeySchema]
    Projection: EncodedProjection


class EncodedGlobalSecondaryIndex(EncodedLocalSecondaryIndex, total=False):
    ProvisionedThroughput: EncodedThroughput


class EncodedStreamSpecificationRequired(TypedDict):
    StreamEnabled: bool


class EncodedStreamSpecification(EncodedStreamSpecificationRequired, total=False):
    StreamViewType: str


SIMPLE_TYPES = frozenset({AttributeType.boolean, AttributeType.string})


NumericTypeConverter = Callable[[str], Any]



from __future__ import annotations

import asyncio
import json
from dataclasses import dataclass
from typing import (
    Any,
    AsyncIterator,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
    cast,
)

from yarl import URL

from aiodynamo.operations import ConditionCheck, Delete, Get, Put, Update

from .credentials import Credentials
from .errors import (
    BrokenThrottleConfig,
    EmptyItem,
    ExpiredToken,
    InternalDynamoError,
    ItemNotFound,
    NoCredentialsFound,
    ProvisionedThroughputExceeded,
    ServiceUnavailable,
    TableDidNotBecomeActive,
    TableDidNotBecomeDisabled,
    TableNotFound,
    Throttled,
    TimeToLiveStatusNotChanged,
    TooManyTransactions,
    TransactionEmpty,
    exception_from_response,
)
from .expressions import (
    Condition,
    KeyCondition,
    Parameters,
    ProjectionExpression,
    UpdateExpression,
)
from .http.types import HttpImplementation, Request, RequestFailed
from .models import (
    BatchGetRequest,
    BatchGetResponse,
    BatchWriteRequest,
    BatchWriteResult,
    GlobalSecondaryIndex,
    KeySchema,
    LocalSecondaryIndex,
    Page,
    PayPerRequest,
    RetryConfig,
    RetryTimeout,
    ReturnValues,
    Select,
    StreamSpecification,
    TableDescription,
    TableStatus,
    Throughput,
    ThroughputType,
    TimeToLiveDescription,
    TimeToLiveStatus,
)
from .sign import signed_dynamo_request
from .types import Item, NumericTypeConverter, TableName
from .utils import dy2py, logger, py2dy, wait


@dataclass(frozen=True)
class TimeToLive:
    table: Table

    async def enable(
        self, attribute: str, *, wait_for_enabled: Union[bool, RetryConfig] = False
    ) -> None:
        await self.table.client.enable_time_to_live(
            self.table.name, attribute, wait_for_enabled=wait_for_enabled
        )

    async def disable(
        self, attribute: str, *, wait_for_disabled: Union[bool, RetryConfig] = False
    ) -> None:
        await self.table.client.disable_time_to_live(
            self.table.name, attribute, wait_for_disabled=wait_for_disabled
        )

    async def describe(self) -> TimeToLiveDescription:
        return await self.table.client.describe_time_to_live(self.table.name)


@dataclass(frozen=True)
class Table:
    client: Client
    name: TableName

    async def exists(self) -> bool:
        return await self.client.table_exists(self.name)

    async def create(
        self,
        throughput: ThroughputType,
        keys: KeySchema,
        *,
        lsis: Optional[List[LocalSecondaryIndex]] = None,
        gsis: Optional[List[GlobalSecondaryIndex]] = None,
        stream: Optional[StreamSpecification] = None,
        wait_for_active: Union[bool, RetryConfig] = False,
    ) -> TableDescription:
        return await self.client.create_table(
            self.name,
            throughput,
            keys,
            lsis=lsis,
            gsis=gsis,
            stream=stream,
            wait_for_active=wait_for_active,
        )

    @property
    def time_to_live(self) -> TimeToLive:
        return TimeToLive(self)

    async def describe(self) -> TableDescription:
        return await self.client.describe_table(self.name)

    async def delete(
        self, *, wait_for_disabled: Union[bool, RetryConfig] = False
    ) -> None:
        return await self.client.delete_table(
            self.name, wait_for_disabled=wait_for_disabled
        )

    async def delete_item(
        self,
        key: Dict[str, Any],
        *,
        return_values: ReturnValues = ReturnValues.none,
        condition: Optional[Condition] = None,
    ) -> Union[None, Item]:
        return await self.client.delete_item(
            self.name, key, return_values=return_values, condition=condition
        )

    async def get_item(
        self,
        key: Dict[str, Any],
        *,
        projection: Optional[ProjectionExpression] = None,
        consistent_read: bool = False,
    ) -> Item:
        """
        Returns the attributes of an item from table.
        This will return all attributes by default.
        To get only some attributes, use a projection expression.
        """
        return await self.client.get_item(
            self.name, key, projection=projection, consistent_read=consistent_read
        )

    async def put_item(
        self,
        item: Dict[str, Any],
        *,
        return_values: ReturnValues = ReturnValues.none,
        condition: Optional[Condition] = None,
    ) -> Union[None, Item]:
        """
        Create a new item or replace it if it already exists.
        This will overwrite all attributes in an item.
        """
        return await self.client.put_item(
            self.name, item, return_values=return_values, condition=condition
        )

    def query(
        self,
        key_condition: KeyCondition,
        *,
        start_key: Optional[Dict[str, Any]] = None,
        filter_expression: Optional[Condition] = None,
        scan_forward: bool = True,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        projection: Optional[ProjectionExpression] = None,
        select: Select = Select.all_attributes,
        consistent_read: bool = False,
    ) -> AsyncIterator[Item]:
        """
        Query the table.

        Unlike the DynamoDB API, the results are automatically de-paginated
        and a single stream of items is returned. For manual pagination, use
        query_single_page(...) instead.

        To filter the result, use a filter expression.
        This will return all attributes by default.
        To get only some attributes, use a projection expression.
        """
        return self.client.query(
            self.name,
            key_condition,
            start_key=start_key,
            filter_expression=filter_expression,
            scan_forward=scan_forward,
            index=index,
            limit=limit,
            projection=projection,
            select=select,
            consistent_read=consistent_read,
        )

    async def query_single_page(
        self,
        key_condition: KeyCondition,
        *,
        start_key: Optional[Dict[str, Any]] = None,
        filter_expression: Optional[Condition] = None,
        scan_forward: bool = True,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        projection: Optional[ProjectionExpression] = None,
        select: Select = Select.all_attributes,
        consistent_read: bool = False,
    ) -> Page:
        """
        Query a single DynamoDB page.
        To automatically handle pagination, uses query(...) instead.
        """
        return await self.client.query_single_page(
            self.name,
            key_condition,
            start_key=start_key,
            filter_expression=filter_expression,
            scan_forward=scan_forward,
            index=index,
            limit=limit,
            projection=projection,
            select=select,
            consistent_read=consistent_read,
        )

    def scan(
        self,
        *,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        start_key: Optional[Dict[str, Any]] = None,
        projection: Optional[ProjectionExpression] = None,
        filter_expression: Optional[Condition] = None,
        consistent_read: bool = False,
    ) -> AsyncIterator[Item]:
        """
        Scan the table.

        Unlike the DynamoDB API, the results are automatically de-paginated
        and a single stream of items is returned. For manual pagination, use
        scan_single_page(...) instead.

        To filter the result, use a filter expression.
        This will return all attributes by default.
        To get only some attributes, use a projection expression.
        """
        return self.client.scan(
            self.name,
            index=index,
            limit=limit,
            start_key=start_key,
            projection=projection,
            filter_expression=filter_expression,
            consistent_read=consistent_read,
        )

    async def scan_single_page(
        self,
        *,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        start_key: Optional[Dict[str, Any]] = None,
        projection: Optional[ProjectionExpression] = None,
        filter_expression: Optional[Condition] = None,
        consistent_read: bool = False,
    ) -> Page:
        """
        Scan a single DynamoDB page.
        To automatically handle pagination, uses scan(...) instead.
        """
        return await self.client.scan_single_page(
            table=self.name,
            index=index,
            limit=limit,
            start_key=start_key,
            projection=projection,
            filter_expression=filter_expression,
            consistent_read=consistent_read,
        )

    async def count(
        self,
        key_condition: KeyCondition,
        *,
        start_key: Optional[Dict[str, Any]] = None,
        filter_expression: Optional[Condition] = None,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        consistent_read: bool = False,
    ) -> int:
        return await self.client.count(
            self.name,
            key_condition,
            start_key=start_key,
            filter_expression=filter_expression,
            index=index,
            limit=limit,
            consistent_read=consistent_read,
        )

    async def scan_count(
        self,
        *,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        start_key: Optional[Dict[str, Any]] = None,
        filter_expression: Optional[Condition] = None,
        consistent_read: bool = False,
    ) -> int:
        return await self.client.scan_count(
            self.name,
            index=index,
            limit=limit,
            start_key=start_key,
            filter_expression=filter_expression,
            consistent_read=consistent_read,
        )

    async def update_item(
        self,
        key: Item,
        update_expression: UpdateExpression,
        *,
        return_values: ReturnValues = ReturnValues.none,
        condition: Optional[Condition] = None,
    ) -> Union[Item, None]:
        """
        Edit an item's attribute or create a new item if it does not exist.
        This will edit only the passed attributes.
        """
        return await self.client.update_item(
            self.name,
            key,
            update_expression,
            return_values=return_values,
            condition=condition,
        )


@dataclass(frozen=True)
class Client:
    http: HttpImplementation
    credentials: Credentials
    region: str
    endpoint: Optional[URL] = None
    numeric_type: NumericTypeConverter = float
    throttle_config: RetryConfig = RetryConfig.default()

    def table(self, name: str) -> Table:
        return Table(self, name)

    async def table_exists(self, name: TableName) -> bool:
        try:
            description = await self.describe_table(name)
        except TableNotFound:
            return False

        return description.status is TableStatus.active

    async def create_table(
        self,
        name: TableName,
        throughput: Union[Throughput, PayPerRequest],
        keys: KeySchema,
        *,
        lsis: Optional[List[LocalSecondaryIndex]] = None,
        gsis: Optional[List[GlobalSecondaryIndex]] = None,
        stream: Optional[StreamSpecification] = None,
        wait_for_active: Union[bool, RetryConfig] = False,
    ) -> TableDescription:
        attributes = keys.to_attributes()
        if lsis is not None:
            for index in lsis:
                attributes.update(index.schema.to_attributes())
        if gsis is not None:
            for index in gsis:
                attributes.update(index.schema.to_attributes())
        attribute_definitions = [
            {"AttributeName": key, "AttributeType": value}
            for key, value in attributes.items()
        ]
        payload: Dict[str, Any] = {
            "AttributeDefinitions": attribute_definitions,
            "TableName": name,
            "KeySchema": keys.encode(),
            **throughput.encode(),
        }
        if lsis:
            payload["LocalSecondaryIndexes"] = [index.encode() for index in lsis]
        if gsis:
            payload["GlobalSecondaryIndexes"] = [index.encode() for index in gsis]
        if stream:
            payload["StreamSpecification"] = stream.encode()

        response = await self.send_request(action="CreateTable", payload=payload)

        description = TableDescription.from_response(response["TableDescription"])

        if description.status == TableStatus.active:
            return description

        async def check() -> bool:
            nonlocal description
            try:
                description = await self.describe_table(name)
                return description.status == TableStatus.active
            except TableNotFound:
                return False

        if not await wait(wait_for_active, check):
            raise TableDidNotBecomeActive()
        return description

    async def enable_time_to_live(
        self,
        table: TableName,
        attribute: str,
        *,
        wait_for_enabled: Union[bool, RetryConfig] = False,
    ) -> None:
        await self.set_time_to_live(
            table, attribute, True, wait_for_change=wait_for_enabled
        )

    async def describe_time_to_live(self, table: TableName) -> TimeToLiveDescription:
        response = await self.send_request(
            action="DescribeTimeToLive", payload={"TableName": table}
        )
        return TimeToLiveDescription(
            table=table,
            attribute=response["TimeToLiveDescription"].get("AttributeName"),
            status=TimeToLiveStatus(
                response["TimeToLiveDescription"]["TimeToLiveStatus"]
            ),
        )

    async def disable_time_to_live(
        self,
        table: TableName,
        attribute: str,
        *,
        wait_for_disabled: Union[bool, RetryConfig] = False,
    ) -> None:
        await self.set_time_to_live(
            table, attribute, False, wait_for_change=wait_for_disabled
        )

    async def set_time_to_live(
        self,
        table: TableName,
        attribute: str,
        status: bool,
        *,
        wait_for_change: Union[bool, RetryConfig] = False,
    ) -> None:
        await self.send_request(
            action="UpdateTimeToLive",
            payload={
                "TableName": table,
                "TimeToLiveSpecification": {
                    "AttributeName": attribute,
                    "Enabled": status,
                },
            },
        )

        desired_result_state = (
            TimeToLiveStatus.enabled if status else TimeToLiveStatus.disabled
        )

        async def check() -> bool:
            description = await self.describe_time_to_live(table)
            return description.status == desired_result_state

        if not await wait(wait_for_change, check):
            raise TimeToLiveStatusNotChanged()

    async def describe_table(self, name: TableName) -> TableDescription:
        response = await self.send_request(
            action="DescribeTable", payload={"TableName": name}
        )

        return TableDescription.from_response(response["Table"])

    async def delete_item(
        self,
        table: str,
        key: Dict[str, Any],
        *,
        return_values: ReturnValues = ReturnValues.none,
        condition: Optional[Condition] = None,
    ) -> Union[None, Item]:
        dynamo_key = py2dy(key)
        if not dynamo_key:
            raise EmptyItem()

        payload: Dict[str, Any] = {
            "TableName": table,
            "Key": dynamo_key,
            "ReturnValues": return_values.value,
        }

        if condition:
            params = Parameters()
            payload["ConditionExpression"] = condition.encode(params)
            payload.update(params.to_request_payload())

        resp = await self.send_request(action="DeleteItem", payload=payload)
        if "Attributes" in resp:
            return dy2py(resp["Attributes"], self.numeric_type)

        else:
            return None

    async def delete_table(
        self,
        table: TableName,
        *,
        wait_for_disabled: Union[bool, RetryConfig] = False,
    ) -> None:
        await self.send_request(action="DeleteTable", payload={"TableName": table})

        async def check() -> bool:
            try:
                await self.describe_table(table)
            except TableNotFound:
                return True
            return False

        if not await wait(wait_for_disabled, check):
            raise TableDidNotBecomeDisabled()

    async def get_item(
        self,
        table: TableName,
        key: Dict[str, Any],
        *,
        projection: Optional[ProjectionExpression] = None,
        consistent_read: bool = False,
    ) -> Item:
        dynamo_key = py2dy(key)
        if not dynamo_key:
            raise EmptyItem()

        payload: Dict[str, Any] = {
            "TableName": table,
            "Key": dynamo_key,
            "ConsistentRead": consistent_read,
        }
        if projection:
            params = Parameters()
            payload["ProjectionExpression"] = projection.encode(params)
            payload.update(params.to_request_payload())

        resp = await self.send_request(action="GetItem", payload=payload)
        if "Item" in resp:
            return dy2py(resp["Item"], self.numeric_type)

        else:
            raise ItemNotFound(dynamo_key)

    async def put_item(
        self,
        table: TableName,
        item: Dict[str, Any],
        *,
        return_values: ReturnValues = ReturnValues.none,
        condition: Optional[Condition] = None,
    ) -> Union[None, Item]:
        dynamo_item = py2dy(item)
        if not dynamo_item:
            raise EmptyItem()
        payload: Dict[str, Any] = {
            "TableName": table,
            "Item": dynamo_item,
            "ReturnValues": return_values.value,
        }
        if condition:
            params = Parameters()
            payload["ConditionExpression"] = condition.encode(params)
            payload.update(params.to_request_payload())

        resp = await self.send_request(action="PutItem", payload=payload)

        if "Attributes" in resp:
            return dy2py(resp["Attributes"], self.numeric_type)

        else:
            return None

    async def query(
        self,
        table: TableName,
        key_condition: KeyCondition,
        *,
        start_key: Optional[Dict[str, Any]] = None,
        filter_expression: Optional[Condition] = None,
        scan_forward: bool = True,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        projection: Optional[ProjectionExpression] = None,
        select: Select = Select.all_attributes,
        consistent_read: bool = False,
    ) -> AsyncIterator[Item]:
        """
        Query the table.
        Unlike the DynamoDB API, the results are automatically de-paginated
        and a single stream of items is returned. For manual pagination, use
        query_single_page(...) instead.
        """
        payload = _query_payload(
            table=table,
            key_condition=key_condition,
            start_key=start_key,
            filter_expression=filter_expression,
            scan_forward=scan_forward,
            index=index,
            projection=projection,
            select=select,
            consistent_read=consistent_read,
        )

        async for result in self._depaginate("Query", payload, limit):
            for item in result["Items"]:
                yield dy2py(item, self.numeric_type)

    async def query_single_page(
        self,
        table: TableName,
        key_condition: KeyCondition,
        *,
        start_key: Optional[Dict[str, Any]] = None,
        filter_expression: Optional[Condition] = None,
        scan_forward: bool = True,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        projection: Optional[ProjectionExpression] = None,
        select: Select = Select.all_attributes,
        consistent_read: bool = False,
    ) -> Page:
        """
        Query a single DynamoDB page.
        To automatically handle pagination, uses query(...) instead.
        """
        payload = _query_payload(
            table=table,
            key_condition=key_condition,
            start_key=start_key,
            filter_expression=filter_expression,
            scan_forward=scan_forward,
            index=index,
            projection=projection,
            select=select,
            consistent_read=consistent_read,
        )
        if limit is not None:
            payload["Limit"] = limit

        response = await self.send_request(action="Query", payload=payload)

        last_evaluated_key: Optional[Dict[str, Any]]
        try:
            last_evaluated_key = dy2py(response["LastEvaluatedKey"], self.numeric_type)
        except KeyError:
            last_evaluated_key = None

        return Page(
            items=[dy2py(item, self.numeric_type) for item in response["Items"]],
            last_evaluated_key=last_evaluated_key,
        )

    async def scan(
        self,
        table: TableName,
        *,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        start_key: Optional[Dict[str, Any]] = None,
        projection: Optional[ProjectionExpression] = None,
        filter_expression: Optional[Condition] = None,
        consistent_read: bool = False,
    ) -> AsyncIterator[Item]:
        """
        Scan the table.
        Unlike the DynamoDB API, the results are automatically de-paginated
        and a single stream of items is returned. For manual pagination, use
        scan_single_page(...) instead.
        """

        payload = _scan_payload(
            table=table,
            index=index,
            start_key=start_key,
            projection=projection,
            filter_expression=filter_expression,
            consistent_read=consistent_read,
        )

        async for result in self._depaginate("Scan", payload, limit):
            for item in result["Items"]:
                yield dy2py(item, self.numeric_type)

    async def scan_single_page(
        self,
        table: TableName,
        *,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        start_key: Optional[Dict[str, Any]] = None,
        projection: Optional[ProjectionExpression] = None,
        filter_expression: Optional[Condition] = None,
        consistent_read: bool = False,
    ) -> Page:
        """
        Scan a single DynamoDB page.
        To automatically handle pagination, uses scan(...) instead.
        """

        payload = _scan_payload(
            table=table,
            index=index,
            start_key=start_key,
            projection=projection,
            filter_expression=filter_expression,
            consistent_read=consistent_read,
        )

        if limit is not None:
            payload["Limit"] = limit

        response = await self.send_request(action="Scan", payload=payload)

        last_evaluated_key: Optional[Dict[str, Any]]
        try:
            last_evaluated_key = dy2py(response["LastEvaluatedKey"], self.numeric_type)
        except KeyError:
            last_evaluated_key = None

        return Page(
            items=[dy2py(item, self.numeric_type) for item in response["Items"]],
            last_evaluated_key=last_evaluated_key,
        )

    async def count(
        self,
        table: TableName,
        key_condition: KeyCondition,
        *,
        start_key: Optional[Dict[str, Any]] = None,
        filter_expression: Optional[Condition] = None,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        consistent_read: bool = False,
    ) -> int:
        params = Parameters()

        payload: Dict[str, Any] = {
            "TableName": table,
            "KeyConditionExpression": key_condition.encode(params),
            "Select": Select.count.value,
            "ConsistentRead": consistent_read,
        }

        if start_key:
            payload["ExclusiveStartKey"] = py2dy(start_key)
        if filter_expression:
            payload["FilterExpression"] = filter_expression.encode(params)
        if index:
            payload["IndexName"] = index
        payload.update(params.to_request_payload())

        count_sum = 0
        async for result in self._depaginate("Query", payload, limit):
            count_sum += result["Count"]
        return count_sum

    async def scan_count(
        self,
        table: TableName,
        *,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        start_key: Optional[Dict[str, Any]] = None,
        filter_expression: Optional[Condition] = None,
        consistent_read: bool = False,
    ) -> int:
        """
        Count the number of items returned by a scan operation.
        """

        payload = _scan_payload(
            table=table,
            index=index,
            start_key=start_key,
            filter_expression=filter_expression,
            projection=None,
            consistent_read=consistent_read,
        )
        payload["Select"] = Select.count.value

        count_sum = 0
        async for result in self._depaginate("Scan", payload, limit):
            count_sum += result["Count"]
        return count_sum

    async def update_item(
        self,
        table: TableName,
        key: Item,
        update_expression: UpdateExpression,
        *,
        return_values: ReturnValues = ReturnValues.none,
        condition: Optional[Condition] = None,
    ) -> Union[Item, None]:
        params = Parameters()

        expression = update_expression.encode(params)
        if not expression:
            raise EmptyItem()

        payload: Dict[str, Any] = {
            "TableName": table,
            "Key": py2dy(key),
            "UpdateExpression": expression,
            "ReturnValues": return_values.value,
        }
        if condition:
            payload["ConditionExpression"] = condition.encode(params)

        payload.update(params.to_request_payload())

        resp = await self.send_request(action="UpdateItem", payload=payload)

        if "Attributes" in resp:
            return dy2py(resp["Attributes"], self.numeric_type)
        else:
            return None

    async def batch_get(
        self, request: Dict[TableName, BatchGetRequest]
    ) -> BatchGetResponse:
        payload = {
            "RequestItems": {
                table: get_request.to_request_payload()
                for table, get_request in request.items()
            }
        }
        response = await self.send_request(action="BatchGetItem", payload=payload)
        return BatchGetResponse(
            items={
                table: [dy2py(item, self.numeric_type) for item in items]
                for table, items in response["Responses"].items()
            },
            unprocessed_keys={
                table: [dy2py(key, self.numeric_type) for key in unprocessed["Keys"]]
                for table, unprocessed in response["UnprocessedKeys"].items()
            },
        )

    async def batch_write(
        self, request: Dict[TableName, BatchWriteRequest]
    ) -> Dict[TableName, BatchWriteResult]:
        payload = {
            "RequestItems": {
                table_name: write_request.to_request_payload()
                for table_name, write_request in request.items()
            }
        }
        response = await self.send_request(action="BatchWriteItem", payload=payload)
        result = {}
        for table, items in response["UnprocessedItems"].items():
            undeleted_keys = []
            unput_items = []
            for item in items:
                try:
                    undeleted_keys.append(
                        dy2py(item["DeleteRequest"]["Key"], self.numeric_type)
                    )
                except KeyError:
                    unput_items.append(
                        dy2py(item["PutRequest"]["Item"], self.numeric_type)
                    )
            result[table] = BatchWriteResult(undeleted_keys, unput_items)
        return result

    async def transact_write_items(
        self,
        items: Sequence[Union[Put, Update, Delete, ConditionCheck]],
        *,
        request_token: Optional[str] = None,
    ) -> None:
        if len(items) == 0:
            raise TransactionEmpty("TransactWriteItems must have at least 1 operation")
        if len(items) > 100:
            raise TooManyTransactions(
                "TransactWriteItems must have a maximum of 100 operations"
            )

        payload = {
            "ClientRequestToken": request_token,
            "TransactItems": [item.to_request_payload() for item in items],
        }
        await self.send_request(action="TransactWriteItems", payload=payload)

    async def transact_get_items(
        self,
        items: List[Get],
    ) -> List[Item]:
        if len(items) == 0:
            raise TransactionEmpty("TransactGetItems must have at least 1 operation")
        if len(items) > 100:
            raise TooManyTransactions(
                "TransactGetItems must have a maximum of 100 operations"
            )

        payload = {
            "TransactItems": [item.to_request_payload() for item in items],
        }
        response = await self.send_request(action="TransactGetItems", payload=payload)

        return [
            dy2py(item["Item"], self.numeric_type) for item in response["Responses"]
        ]

    async def send_request(
        self,
        *,
        action: str,
        payload: Mapping[str, Any],
    ) -> Dict[str, Any]:
        """
        Send a request to DynamoDB and handle retries if necessary.

        The self.throttle_config.attempts() async iterable handles retries
        by yielding each time we should do another attempt and raising
        RetryTimeout once the time limit is reached.

        In each iteration of the loop, we send a request to DynamoDB and check
        its result. If it's good, we break out of the loop and return the parsed
        JSON data to the caller.
        If a non-200 status is returned, we determine the error type via
        exception_from_response and check if we should retry. If we should retry,
        log the error and go to the next iteration, otherwise raise the exception.
        In either case, we store the last exception found so if we hit the retry
        limit, we raise that exception.
        If the loop never executed, we raise a BrokenThrottleConfig because
        RetryConfig.attempts() should always yield at least once.
        """
        exception: Optional[Exception] = None
        try:
            async for _ in self.throttle_config.attempts():
                key = await self.credentials.get_key(self.http)
                if key is None:
                    logger.debug("no credentials found")
                    exception = NoCredentialsFound()
                    continue
                request = signed_dynamo_request(
                    key=key,
                    payload=payload,
                    action=action,
                    region=self.region,
                    endpoint=self.endpoint,
                )
                logger.debug("sending request %r", request)
                try:
                    response = await self.http(
                        Request(
                            method="POST",
                            url=str(request.url),
                            headers=request.headers,
                            body=request.body,
                        )
                    )
                except asyncio.TimeoutError as exc:
                    logger.debug("http timeout")
                    exception = exc
                    continue
                except RequestFailed as exc:
                    logger.debug("request failed")
                    exception = exc.inner
                    continue
                logger.debug("got response %r", response)
                if response.status == 200:
                    return cast(Dict[str, Any], json.loads(response.body))
                exception = exception_from_response(response.status, response.body)
                if isinstance(exception, Throttled):
                    logger.debug("request throttled")
                elif isinstance(exception, ProvisionedThroughputExceeded):
                    logger.debug("provisioned throughput exceeded")
                elif isinstance(exception, ExpiredToken):
                    logger.debug("token expired")
                    if not self.credentials.invalidate():
                        raise exception
                elif isinstance(exception, ServiceUnavailable):
                    logger.debug("service unavailable")
                elif isinstance(exception, InternalDynamoError):
                    logger.debug("internal dynamo error")
                else:
                    raise exception
        except RetryTimeout:
            if exception is not None:
                raise exception
            raise
        raise BrokenThrottleConfig()

    async def _depaginate(
        self, action: str, payload: Dict[str, Any], limit: Optional[int] = None
    ) -> AsyncIterator[Dict[str, Any]]:
        """
        Internal API to depaginate the results from query/scan/count.
        Don't call this directly, use .query, .scan or .count instead.
        """
        if limit is not None:
            payload = {**payload, "Limit": limit}
        task: Optional[asyncio.Task[Dict[str, Any]]] = asyncio.create_task(
            self.send_request(action=action, payload=payload)
        )
        is_count = payload.get("Select") == "COUNT"
        try:
            while task:
                result = await task
                try:
                    payload = {
                        **payload,
                        "ExclusiveStartKey": result["LastEvaluatedKey"],
                    }
                except KeyError:
                    task = None
                else:
                    if limit is not None:
                        consumed: int = (
                            result["Count"] if is_count else len(result["Items"])
                        )
                        limit -= consumed
                        if limit > 0:
                            payload["Limit"] = limit
                        else:
                            task = None
                if task:
                    task = asyncio.create_task(
                        self.send_request(action=action, payload=payload)
                    )
                yield result
        except asyncio.CancelledError:
            if task:
                task.cancel()
            raise


def _query_payload(
    *,
    table: TableName,
    key_condition: KeyCondition,
    start_key: Optional[Dict[str, Any]],
    filter_expression: Optional[Condition],
    scan_forward: bool,
    index: Optional[str],
    projection: Optional[ProjectionExpression],
    select: Select = Select.all_attributes,
    consistent_read: bool = False,
) -> Dict[str, Any]:
    if projection:
        select = Select.specific_attributes
    if select is Select.count:
        raise TypeError("Cannot use Select.count with query, use count instead")

    params = Parameters()

    payload: Dict[str, Any] = {
        "TableName": table,
        "KeyConditionExpression": key_condition.encode(params),
        "ScanIndexForward": scan_forward,
        "ConsistentRead": consistent_read,
    }

    if projection:
        payload["ProjectionExpression"] = projection.encode(params)

    if filter_expression:
        payload["FilterExpression"] = filter_expression.encode(params)

    if start_key:
        payload["ExclusiveStartKey"] = py2dy(start_key)
    if index:
        payload["IndexName"] = index
    if select:
        payload["Select"] = select.value

    payload.update(params.to_request_payload())
    return payload


def _scan_payload(
    *,
    table: TableName,
    index: Optional[str],
    start_key: Optional[Dict[str, Any]],
    projection: Optional[ProjectionExpression],
    filter_expression: Optional[Condition],
    consistent_read: bool = False,
) -> Dict[str, Any]:
    params = Parameters()

    payload: Dict[str, Any] = {
        "TableName": table,
        "ConsistentRead": consistent_read,
    }

    if index:
        payload["IndexName"] = index
    if start_key:
        payload["ExclusiveStartKey"] = py2dy(start_key)
    if projection:
        payload["ProjectionExpression"] = projection.encode(params)
    if filter_expression:
        payload["FilterExpression"] = filter_expression.encode(params)

    payload.update(params.to_request_payload())
    return payload



from dataclasses import dataclass
from typing import Awaitable, Callable, Dict, Literal, Optional, Union


@dataclass(frozen=True)
class Request:
    method: Union[Literal["GET"], Literal["POST"], Literal["PUT"]]
    url: str
    headers: Optional[Dict[str, str]]
    body: Optional[bytes]


@dataclass(frozen=True)
class Response:
    status: int
    body: bytes


@dataclass
class RequestFailed(Exception):
    inner: Exception


HttpImplementation = Callable[[Request], Awaitable[Response]]



import abc
from dataclasses import dataclass
from typing import Any, Dict, Optional

from aiodynamo.errors import EmptyItem
from aiodynamo.expressions import (
    Condition,
    Parameters,
    ProjectionExpression,
    UpdateExpression,
)
from aiodynamo.types import Item, TableName
from aiodynamo.utils import py2dy


class Operation(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def to_request_payload(self) -> Dict[str, Any]:
        pass


@dataclass(frozen=True)
class Get(Operation):
    table: TableName
    key: Item
    projection: Optional[ProjectionExpression] = None

    def to_request_payload(self) -> Dict[str, Any]:
        dynamo_key = py2dy(self.key)
        if not dynamo_key:
            raise EmptyItem()

        payload: Dict[str, Any] = {
            "TableName": self.table,
            "Key": dynamo_key,
        }

        if self.projection:
            params = Parameters()
            payload["ProjectionExpression"] = self.projection.encode(params)
            payload.update(params.to_request_payload())

        return {"Get": payload}


@dataclass(frozen=True)
class Put(Operation):
    table: TableName
    item: Item
    condition: Optional[Condition] = None

    def to_request_payload(self) -> Dict[str, Any]:
        dynamo_item = py2dy(self.item)
        if not dynamo_item:
            raise EmptyItem()

        payload: Dict[str, Any] = {
            "TableName": self.table,
            "Item": dynamo_item,
        }
        if self.condition:
            params = Parameters()
            payload["ConditionExpression"] = self.condition.encode(params)
            payload.update(params.to_request_payload())

        return {"Put": payload}


@dataclass(frozen=True)
class Update(Operation):
    table: TableName
    key: Item
    expression: UpdateExpression
    condition: Optional[Condition] = None

    def to_request_payload(self) -> Dict[str, Any]:
        params = Parameters()
        expression = self.expression.encode(params)
        if not expression:
            raise EmptyItem()

        payload: Dict[str, Any] = {
            "TableName": self.table,
            "UpdateExpression": expression,
            "Key": py2dy(self.key),
        }

        if self.condition:
            payload["ConditionExpression"] = self.condition.encode(params)

        payload.update(params.to_request_payload())

        return {"Update": payload}


@dataclass(frozen=True)
class Delete(Operation):
    table: TableName
    key: Item
    condition: Optional[Condition] = None

    def to_request_payload(self) -> Dict[str, Any]:
        dynamo_item = py2dy(self.key)
        if not dynamo_item:
            raise EmptyItem()

        payload: Dict[str, Any] = {
            "TableName": self.table,
            "Key": dynamo_item,
        }
        if self.condition:
            params = Parameters()
            payload["ConditionExpression"] = self.condition.encode(params)
            payload.update(params.to_request_payload())

        return {"Delete": payload}


@dataclass(frozen=True)
class ConditionCheck(Operation):
    table: TableName
    key: Item
    condition: Optional[Condition] = None

    def to_request_payload(self) -> Dict[str, Any]:
        dynamo_item = py2dy(self.key)
        if not dynamo_item:
            raise EmptyItem()

        payload: Dict[str, Any] = {
            "TableName": self.table,
            "Key": dynamo_item,
        }
        if self.condition:
            params = Parameters()
            payload["ConditionExpression"] = self.condition.encode(params)
            payload.update(params.to_request_payload())

        return {"ConditionCheck": payload}



import json
from dataclasses import dataclass
from typing import Any, Dict, List, Union


@dataclass(frozen=True)
class CancellationReason:
    code: str
    message: str


class AIODynamoError(Exception):
    pass


class NoCredentialsFound(AIODynamoError):
    pass


class EmptyItem(AIODynamoError):
    pass


class ItemNotFound(AIODynamoError):
    pass


class CannotAddToNestedField(AIODynamoError):
    pass


class UnknownError(AIODynamoError):
    def __init__(self, status: int, body: bytes):
        self.status = status
        self.body = body
        super().__init__(body)


class UnknownOperation(AIODynamoError):
    pass


class TableNotFound(AIODynamoError):
    pass


class ProvisionedThroughputExceeded(AIODynamoError):
    pass


class RequestLimitExceeded(AIODynamoError):
    pass


class InternalDynamoError(AIODynamoError):
    pass


class ItemCollectionSizeLimitExceeded(AIODynamoError):
    pass


class BackupInUse(AIODynamoError):
    pass


class ContinuousBackupsUnavailable(AIODynamoError):
    pass


class TableInUse(AIODynamoError):
    pass


class ResourceInUse(AIODynamoError):
    pass


class GlobalTableAlreadyExists(AIODynamoError):
    pass


class SourceTableDoesNotExist(AIODynamoError):
    pass


class LimitExceeded(AIODynamoError):
    pass


class BackupNotFound(AIODynamoError):
    pass


class ConditionalCheckFailed(AIODynamoError):
    pass


class TransactionConflict(AIODynamoError):
    pass


class GlobalTableNotFound(AIODynamoError):
    pass


class TableAlreadyExists(AIODynamoError):
    pass


class InvalidRestoreTime(AIODynamoError):
    pass


class PointInTimeRecoveryUnavailable(AIODynamoError):
    pass


class TransactionCanceled(AIODynamoError):
    cancellation_reasons: List[Union[CancellationReason, None]]

    def __init__(self, body: Dict[str, Any]):
        self.body = body
        self.cancellation_reasons: List[CancellationReason] = [
            CancellationReason(reason["Code"], reason["Message"])
            if reason["Code"] != "None"
            else None
            for reason in self.body["CancellationReasons"]
        ]
        super().__init__(body)


class TransactionEmpty(AIODynamoError):
    pass


class TooManyTransactions(AIODynamoError):
    pass


class ReplicaAlreadyExists(AIODynamoError):
    pass


class ReplicaNotFound(AIODynamoError):
    pass


class TableDidNotBecomeActive(AIODynamoError):
    pass


class TableDidNotBecomeDisabled(AIODynamoError):
    pass


class Throttled(AIODynamoError):
    pass


class BrokenThrottleConfig(Throttled):
    pass


class ValidationException(AIODynamoError):
    pass


class TimeToLiveStatusNotChanged(AIODynamoError):
    pass


class ExpiredToken(AIODynamoError):
    pass


class ServiceUnavailable(AIODynamoError):
    pass


class IdempotentParameterMismatch(AIODynamoError):
    pass


class TransactionInProgress(AIODynamoError):
    pass


ERRORS = {
    "ResourceNotFoundException": TableNotFound,
    "UnknownOperationException": UnknownOperation,
    "ProvisionedThroughputExceededException": ProvisionedThroughputExceeded,
    "RequestLimitExceeded": RequestLimitExceeded,
    "ItemCollectionSizeLimitExceededException": ItemCollectionSizeLimitExceeded,
    "BackupInUseException": BackupInUse,
    "ContinuousBackupsUnavailableException": ContinuousBackupsUnavailable,
    "TableInUseException": TableInUse,
    "GlobalTableAlreadyExistsException": GlobalTableAlreadyExists,
    "TableNotFoundException": SourceTableDoesNotExist,
    "LimitExceededException": LimitExceeded,
    "BackupNotFoundException": BackupNotFound,
    "ConditionalCheckFailedException": ConditionalCheckFailed,
    "TransactionConflictException": TransactionConflict,
    "GlobalTableNotFoundException": GlobalTableNotFound,
    "TableAlreadyExistsException": TableAlreadyExists,
    "InvalidRestoreTimeException": InvalidRestoreTime,
    "PointInTimeRecoveryUnavailableException": PointInTimeRecoveryUnavailable,
    "TransactionCanceledException": TransactionCanceled,
    "ReplicaAlreadyExistsException": ReplicaAlreadyExists,
    "ReplicaNotFoundException": ReplicaNotFound,
    "ThrottlingException": Throttled,
    "ValidationException": ValidationException,
    "ExpiredTokenException": ExpiredToken,
    "ResourceInUseException": ResourceInUse,
    "IdempotentParameterMismatchException": IdempotentParameterMismatch,
    "TransactionInProgressException": TransactionInProgress,
}


def exception_from_response(status: int, body: bytes) -> Exception:
    if status == 500:
        return InternalDynamoError()
    elif status == 503:
        return ServiceUnavailable()
    try:
        data = json.loads(body)
        error: Exception = ERRORS[data["__type"].split("#", 1)[-1]](data)
        return error
    except Exception:
        return UnknownError(status, body)
